from .kurbopy import Point as Point
from .kurbopy import Rect as Rect
from .kurbopy import Vec2 as Vec2
from .kurbopy import Line as Line
from .kurbopy import CubicBez as CubicBez
from .kurbopy import QuadBez as QuadBez
from .kurbopy import TranslateScale as TranslateScale
from .kurbopy import BezPath as BezPath
from fontTools.pens.basePen import BasePen
from kurbopy.magic import magic_mul, magic_add, magic_sub
import re


def fromDrawable(drawable, *penArgs, **penKwargs):
    """Returns an *array of BezPath* from any object conforming to the pen protocol."""
    pen = BezPathCreatingPen(*penArgs, **penKwargs)
    drawable.draw(pen)
    return pen.paths


setattr(BezPath, "fromDrawable", fromDrawable)


def to_matplot(self):
    from matplotlib.path import Path

    svg = self.to_svg()
    verts = []
    codes = []
    while len(svg):
        m = re.match(r"^M(-?[\d\.]+) (-?[\d\.]+)\s*", svg)
        if m:
            codes.append(Path.MOVETO)
            verts.append((m[1], m[2]))
            svg = svg[len(m[0]) :]
            continue

        m = re.match(r"^L(-?[\d\.]+) (-?[\d\.]+)\s*", svg)
        if m:
            codes.append(Path.LINETO)
            verts.append((m[1], m[2]))
            svg = svg[len(m[0]) :]
            continue

        m = re.match(r"^Q(-?[\d\.]+) (-?[\d\.]+) (-?[\d\.]+) (-?[\d\.]+)\s*", svg)
        if m:
            codes.append(Path.CURVE3)
            verts.append((m[1], m[2]))
            codes.append(Path.CURVE3)
            verts.append((m[3], m[4]))
            svg = svg[len(m[0]) :]
            continue

        m = re.match(
            r"^C(-?[\d\.]+) (-?[\d\.]+) (-?[\d\.]+) (-?[\d\.]+) (-?[\d\.]+) (-?[\d\.]+)\s*",
            svg,
        )
        if m:
            codes.append(Path.CURVE4)
            verts.append((m[1], m[2]))
            codes.append(Path.CURVE4)
            verts.append((m[3], m[4]))
            codes.append(Path.CURVE4)
            verts.append((m[5], m[6]))
            svg = svg[len(m[0]) :]
            continue

        if svg[0] == "Z":
            verts.append(verts[0])
            codes.append(Path.CLOSEPOLY)
            svg = svg[1:]
            continue
        raise (ValueError("Don't know what to do with " + svg))

    return Path(verts, codes)


setattr(BezPath, "to_matplot", to_matplot)


def plot(self, ax, **kwargs):
    """Plot the path on a Matplot subplot which you supply

    ::

          import matplotlib.pyplot as plt
          fig, ax = plt.subplots()
          path.plot(ax)

    """
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    from matplotlib.path import Path
    import matplotlib.patches as patches

    path = self.to_matplot()
    if not "lw" in kwargs:
        kwargs["lw"] = 2
    if not "fill" in kwargs:
        kwargs["fill"] = False
    drawNodes = not ("drawNodes" in kwargs) or kwargs["drawNodes"] != False
    if "drawNodes" in kwargs:
        kwargs.pop("drawNodes")
    patch = patches.PathPatch(path, **kwargs)
    ax.add_patch(patch)
    left, right = ax.get_xlim()
    top, bottom = ax.get_ylim()
    bounds = self.bounding_box()
    bounds = bounds.inset(50)
    if not (left == 0.0 and right == 1.0 and top == 0.0 and bottom == 1.0):
        bounds = bounds.union_pt(Point(left, top))
        bounds = bounds.union_pt(Point(right, bottom))
    ax.set_xlim(bounds.min_x(), bounds.max_x())
    ax.set_ylim(bounds.min_y(), bounds.max_y())


setattr(BezPath, "plot", plot)


class BezPathCreatingPen(BasePen):
    def __init__(self, *args, **kwargs):
        super(BezPathCreatingPen, self).__init__(*args, **kwargs)
        self.paths = []
        self.path = BezPath()

    def _moveTo(self, p):
        self.path.move_to(Point(p[0], p[1]))

    def _lineTo(self, p):
        self.path.line_to(Point(p[0], p[1]))

    def _curveToOne(self, p1, p2, p3):
        self.path.curve_to(
            Point(p1[0], p1[1]), Point(p2[0], p2[1]), Point(p3[0], p3[1])
        )

    def _qCurveToOne(self, p1, p2):
        self.path.quad_to(Point(p1[0], p1[1]), Point(p2[0], p2[1]))

    def _closePath(self):
        self.path.close_path()
        self.paths.append(self.path)
        self.path = BezPath()




# This is dirty
def get_magic_name(obj):
    cls = obj.__class__
    if hasattr(cls, "__name__"):
        return cls.__name__
    print(str(obj))
    sys.exit(1)


def magic_mul(self, rhs):
    other_type = get_magic_name(rhs)
    mytype = get_magic_name(self)
    if hasattr(self, "_mul_" + other_type):
        return getattr(self, "_mul_" + other_type)(rhs)
    raise TypeError("Cannot multiply %s by %s" % (mytype, other_type))


def magic_sub(self, rhs):
    other_type = get_magic_name(rhs.__class__)
    if hasattr(self, "_sub_" + other_type):
        return getattr(self, "_sub_" + other_type)(rhs)
    raise TypeError("Cannot subtract %s from %s" % (other_type, mytype))

def magic_isub(self, rhs):
    other_type = get_magic_name(rhs.__class__)
    if hasattr(self, "_isub_" + other_type):
        return getattr(self, "_isub_" + other_type)(rhs)
    raise TypeError("Cannot subtract %s from %s" % (other_type, mytype))

def magic_add(self, rhs):
    other_type = get_magic_name(rhs.__class__)
    if hasattr(self, "_add_" + other_type):
        return getattr(self, "_add_" + other_type)(rhs)
    raise TypeError("Cannot add %s to %s" % (other_type, mytype))

def magic_iadd(self, rhs):
    other_type = get_magic_name(rhs.__class__)
    if hasattr(self, "_iadd_" + other_type):
        return getattr(self, "_iadd_" + other_type)(rhs)
    raise TypeError("Cannot add %s to %s" % (other_type, mytype))



from babelfont import load
from kurbopy import BezPath, TranslateScale, Point, Vec2
from glyphtools import get_glyph_metrics, get_beziers
import logging

logger = logging.getLogger()


min_bubble = 100
font = f = load("/Users/simon/work/nastaliq/gulzar/sources/Gulzar.glyphs")
glyphset = {k: font.default_master.get_glyph_layer(k) for k in font.exportedGlyphs()}
glyphcache = {}


def get_beziers_new(font, glyph):
    if glyph in glyphcache:
        return glyphcache[glyph]
    layer = font.default_master.get_glyph_layer(glyph)
    rv = BezPath.fromDrawable(layer, glyphset)
    glyphcache[glyph] = rv
    return rv


def determine_kern_new(
    font, glyph1, glyph2, targetdistance, offset1=(0, 0), offset2=(0, 0), maxtuck=0.4,
):
    paths1 = get_beziers_new(font, glyph1)
    paths2 = get_beziers_new(font, glyph2)
    metrics1 = get_glyph_metrics(font, glyph1)
    metrics2 = get_glyph_metrics(font, glyph2)

    offset1 = TranslateScale.translate(Vec2(*offset1))
    offset2 = Vec2(offset2[0] + metrics1["width"], offset2[1])
    kern = 0
    last_best = None

    minimum_possible = -1000
    full_width = max(metrics1["run"] - min(0, metrics1["rsb"]), min_bubble)
    logger.debug("Full width of %s is %i" % (glyph1, full_width))
    if maxtuck:
        maximum_width = full_width * maxtuck
        logger.debug("Maximum distance into %s is %i" % (glyph1, maximum_width))
        logger.debug("Left edge of %s is %i" % (glyph2, metrics2["xMin"]))
        minimum_possible = -metrics2["xMin"] - maximum_width
        logger.debug("Biggest kern is %i" % minimum_possible)

    iterations = 0
    while True:
        # Compute min distance
        min_distance = None
        for p1 in paths1:
            moved_p1 = offset1 * p1
            for p2 in paths2:
                t2 = TranslateScale.translate(offset2 + Vec2(kern, 0))
                moved_p2 = t2 * p2
                d = moved_p1.min_distance(moved_p2)
                # d = cached_distance(p1, p2)
                #                 logger.debug("d was", d)
                # logger.debug("offset1 was %s" % offset1)
                # logger.debug("offset2 was %s" % offset2)
                if not min_distance or d < min_distance:
                    min_distance = d
        if not last_best or min_distance < last_best:
            last_best = min_distance
        else:
            break  # Nothing helped
        if abs(min_distance - targetdistance) < 1 or iterations > 10:
            break
        iterations = iterations + 1
        kern = kern + (targetdistance - min_distance)
        if kern < minimum_possible:
            return minimum_possible
    #     kern = kern -  metrics1["rsb"]
    if maxtuck:
        kern = max(kern, -(metrics1["xMax"] * (1 + maxtuck)) + metrics1["rsb"])
    else:
        kern = max(kern, -(metrics1["xMax"]) + metrics1["rsb"])
    if metrics1["rsb"] < 0:
        kern = kern - metrics1["rsb"]
    # logger.debug("%s/%s/%i = %i" % (glyph1, glyph2, offset2.y, kern))
    return kern  # int(kern)


def determine_kern_old(
    font, glyph1, glyph2, targetdistance, offset1=(0, 0), offset2=(0, 0), maxtuck=0.4,
):
    """Determine a kerning value required to set two glyphs at given ink-to-ink distance.

    The value is bounded by the ``maxtuck`` parameter. For example, if
    ``maxtuck`` is 0.20, the right glyph will not be placed any further
    left than 80% of the width of left glyph, even if this places the
    ink further than ``targetdistance`` units away.

    Args:
        font: a ``fontTools`` TTFont object or a ``glyphsLib`` GSFontMaster
              object OR a ``babelfont`` Font object.
        glyph1: name of the left glyph.
        glyph2: name of the right glyph.
        targetdistance: distance to set the glyphs apart.
        offset1: offset (X-coordinate, Y-coordinate) to place left glyph.
        offset2: offset (X-coordinate, Y-coordinate) to place right glyph.
        maxtuck: maximum proportion of the left glyph's width to kern.

    Returns: A kerning value, in units.
    """

    def cached_distance(p1, p2):
        if (p1, p2) not in distance_cache:
            distance_cache[(p1, p2)] = p1.distanceToPath(p2, samples=3)
        return distance_cache[(p1, p2)]

    paths1 = get_beziers(font, glyph1)
    paths2 = get_beziers(font, glyph2)
    metrics1 = get_glyph_metrics(font, glyph1)
    metrics2 = get_glyph_metrics(font, glyph2)

    offset1 = Point(*offset1)
    offset2 = Point(offset2[0] + metrics1["width"], offset2[1])
    kern = 0
    last_best = None

    minimum_possible = -1000
    full_width = max(metrics1["run"] - min(0, metrics1["rsb"]), min_bubble)
    logger.debug("Full width of %s is %i" % (glyph1, full_width))
    if maxtuck:
        maximum_width = full_width * maxtuck
        logger.debug("Maximum distance into %s is %i" % (glyph1, maximum_width))
        logger.debug("Left edge of %s is %i" % (glyph2, metrics2["xMin"]))
        minimum_possible = -metrics2["xMin"] - maximum_width
        logger.debug("Biggest kern is %i" % minimum_possible)

    iterations = 0
    while True:
        # Compute min distance
        min_distance = None
        for p1 in paths1:
            p1 = p1.clone().translate(offset1)
            for p2 in paths2:
                p2 = p2.clone().translate(Point(offset2.x + kern, offset2.y))
                d = p1.distanceToPath(p2, samples=3)
                # d = cached_distance(p1, p2)
                #                 logger.debug("d was", d)
                # logger.debug("offset1 was %s" % offset1)
                # logger.debug("offset2 was %s" % offset2)
                if not min_distance or d[0] < min_distance:
                    min_distance = d[0]
        if not last_best or min_distance < last_best:
            last_best = min_distance
        else:
            break  # Nothing helped
        if abs(min_distance - targetdistance) < 1 or iterations > 10:
            break
        iterations = iterations + 1
        kern = kern + (targetdistance - min_distance)
        if kern < minimum_possible:
            return minimum_possible
    #     kern = kern -  metrics1["rsb"]
    if maxtuck:
        kern = max(kern, -(metrics1["xMax"] * (1 + maxtuck)) + metrics1["rsb"])
    else:
        kern = max(kern, -(metrics1["xMax"]) + metrics1["rsb"])
    if metrics1["rsb"] < 0:
        kern = kern - metrics1["rsb"]
    logger.debug("%s/%s/%i = %i" % (glyph1, glyph2, offset2.y, kern))
    return int(kern)


def old():
    for i in range(100, 200, 20):
        determine_kern_old(font, "ALIFf1", "LAMi11", i, (0, 0), (0, 200))


def new():
    for i in range(100, 200, 20):
        determine_kern_new(font, "ALIFf1", "LAMi11", i, (0, 0), (0, 200))


i = 150
print(determine_kern_old(font, "ALIFf1", "LAMi11", i, (0, 0), (0, 200)))
print(determine_kern_new(font, "ALIFf1", "LAMi11", i, (0, 0), (0, 200)))
import timeit

print("Old", timeit.timeit(old, number=100))
print("New", timeit.timeit(new, number=100))



from setuptools import setup

version, section = None, None
with open("Cargo.toml") as cargo_toml:
    for line in cargo_toml:
        if line.startswith('['):
            section = line[1:line.index(']')]
        elif section == "package" and line.startswith("version"):
            version = line.split("=")[1].lstrip().rstrip()
            version = version.replace("\"", "")
if not version:
    raise ValueError("No package version string in `Cargo.toml`.")

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

setup(
    name="kurbopy",
    version=version,
    author="Simon Cozens",
    author_email="simon@simon-cozens.org",
    description="A Python wrapper around `kurbo.rs`.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/simoncozens/kurbopy",
    packages=["kurbopy"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Rust",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: MacOS",
    ],
    python_requires='>=3.6',
)
